/*
 * Distributed under the OSI-approved Apache License, Version 2.0.  See
 * accompanying file Copyright.txt for details.
 *
 * adios2_c.h : ADIOS2 C bindings declarations
 *
 *  Created on: Mar 13, 2017
 *      Author: William F Godoy godoywf@ornl.gov
 */

#ifndef ADIOS2_BINDINGS_C_ADIOS2_C_H_
#define ADIOS2_BINDINGS_C_ADIOS2_C_H_

#include "adios2/common/ADIOSConfig.h"

#include "adios2/c/adios2_c_adios.h"
#include "adios2/c/adios2_c_attribute.h"
#include "adios2/c/adios2_c_engine.h"
#include "adios2/c/adios2_c_io.h"
#include "adios2/c/adios2_c_operator.h"
#include "adios2/c/adios2_c_variable.h"

#endif /* ADIOS2_BINDINGS_C_ADIOS2_C_H_ */
