!
! Distributed under the OSI-approved Apache License, Version 2.0.  See
!  accompanying file Copyright.txt for details.
!
!  adios2_mod.f90 : ADIOS2 Fortran bindings central module
!
!   Created on: Mar 13, 2017
!       Author: William F Godoy godoywf@ornl.gov
!

module adios2

    use adios2_parameters_mod
    use adios2_adios_mod
    use adios2_attribute_mod
    use adios2_attribute_data_mod
    use adios2_io_mod
    use adios2_variable_mod
    use adios2_variable_min_mod
    use adios2_variable_max_mod
    use adios2_engine_mod

end module
