/*
 * Distributed under the OSI-approved Apache License, Version 2.0.  See
 * accompanying file Copyright.txt for details.
 *
 * adios2_f2c_common.h
 *
 *  Created on: Sep 18, 2018
 *      Author: William F Godoy godoywf@ornl.gov
 */

#ifndef ADIOS2_BINDINGS_FORTRAN_F2C_ADIOS2_F2C_COMMON_H_
#define ADIOS2_BINDINGS_FORTRAN_F2C_ADIOS2_F2C_COMMON_H_

#include "adios2/common/ADIOSConfig.h"
#include <FC.h>
#include <adios2_c.h>

#include <stdint.h> // int64_t
#include <vector>

#endif /* ADIOS2_BINDINGS_FORTRAN_F2C_ADIOS2_F2C_COMMON_H_ */
