/*
 * Distributed under the OSI-approved Apache License, Version 2.0.  See
 * accompanying file Copyright.txt for details.
 *
 * adiosMemory.tcc
 *
 *  Created on: Jan 23, 2019
 *      Author: William F Godoy godoywf@ornl.gov
 */

#ifndef ADIOS2_HELPER_ADIOSMEMORY_TCC_
#define ADIOS2_HELPER_ADIOSMEMORY_TCC_

#include "adiosMemory.h"

#include "adios2/common/ADIOSMacros.h"

namespace adios2
{
namespace helper
{} // end namespace helper
} // end namespace adios2

#endif /* ADIOS2_HELPER_ADIOSMEMORY_TCC_ */
